;nyquist plug-in
;version 4
;name "Import Sony Track Marks"
;release 0.0.2
;author "Tom Kirchner"
;copyright "Released under terms of the GNU General Public License version 2"
;type tool analyze
;preview disabled
;debugbutton true

;control filepaths "Select one or more track marker files" file "Select" "*default*/YYMMDD_HHMM.tmk" "Sony track marks (*.tmk)|*.tmk;*.TMK|All files|*.*;*" "open,exists,multiple"

;; Tested with Audacity 3.4.2.
;; Based on https://plugins.audacityteam.org/contributing/developing-your-own-plugins-and-scripts/creating-your-own-nyquist-plugins/tutorials/file-button-tutorial#reading_multiple_files
;;
;; This plugin imports track markers from a a .tmk file, as generated by Sony
;; voice recorders when hitting their "T-mark" button.
;;
;; These files start with a binary format identifier, then each line of the file
;; has a timestamp of the form "[00000:02.23]" which is "minutes:seconds".
;;
;; Each track mark is added as an Audacity label.
;;
;; NOTE: The markers are relative to the start of the file, so you should select
;; the entire file before importing.

(defun get-files(txt)
  ; Convert string of file paths to LISP list.
  ; Return list.
  (let* ((txt (string-trim "\"" txt))
         (txt (format nil "(list ~s )" txt)))
    (eval-string txt)))


(defun label-from-text(txt)
  ; 'txt' is a line from the file. It must be formatted as: "[number:number]"
  ; The output needs to be a list in the form: (list number string)
  ;
  ; We build up a string of just the input numbers, later turned into floats
  ; with eval-string.
  (setf newtxt "")
  ; Track parsing state; 'isnumber' is whether we're currently in the middle of
  ; a (real) number; 'index' is which number we're reading.
  (let ((isnumber nil)
        (index 0))
    ; Read each character, skipping the final newline.
    (dotimes (i (- (length txt) 1))
      (setf ch (char txt i))
      (cond
       ((and (< index 2) ; two numbers expected.
             ; Digits or '.' are part of a number.
             (or (char= ch #\.)
                 (digit-char-p ch)))
            ; In a number; append character to string.
            (setf isnumber t)
            (setf newtxt (format nil "~a~a" newtxt ch)))
       ; Minutes and seconds are separated by a colon; this also indicates a
       ; new number is coming.
       ((and isnumber (char= ch #\:))
           (incf index)
           (setf isnumber nil)
           (setf newtxt (format nil "~a " newtxt)))
       ; A closing bracket signals the end of the second number.
       ((and isnumber (char= ch #\]))
           (incf index)
           (setf isnumber nil))
       ; Ignore starting delimiter, the open bracket.
       ((char= ch #\[) ())
       ; Nothing else is allowed
       (t (throw 'err (format nil "Invalid character '~a'; expecting [number:number]" ch))))))

  ; Convert newtxt to a list with (real) numbers.
  (setf numbers (eval-string (format nil "(list ~a)" newtxt)))
  ; Do basic checks on the list.
  (cond
   ((/= (length numbers) 2)
      (throw 'err "Error.  Each line must have (only) minutes and seconds."))
   ((or (not (numberp (first numbers)))
        (not (numberp (second numbers))))
      (throw 'err "Error.  Values are not numbers.")))
  ; Convert minutes and seconds to total number of seconds.
  (setf totaltime (+
                (* 60 (first numbers))
                (second numbers)))

  ; This format tells Audacity it's a label.  We don't have any relevant text.
  (setf data (list totaltime ""))
  data)


; Read a byte from a stream, check that it matches an expected value, and throw it away
(defun checkbyte(fp expected)
  (let ((byte (read-byte fp)))
     (when (not (eql byte expected))
       (throw 'err (format nil "Invalid header byte '~A' expected '~A'" byte expected)))))

; Open file with error checking.
(defun openfile(filename)
  (setf fp (open-binary filename))
  (when (not fp)
    (throw 'err (format nil "Error.~%~s could not be opened." filename)))

  ; Check for (and throw away) magic bytes in Sony track marker format.
  (checkbyte fp #xef) ; 239
  (checkbyte fp #xbb) ; 187
  (checkbyte fp #xbf) ; 191

  fp)


(defun process-file (fp)
  ; Reads data from file pointer and create one label from each line.
  ; Return a list of labels.
  (let ((labels ()))
    (do ((txt (read-line fp) (read-line fp)))
        ((not txt) labels)
      (let ((label (label-from-text txt)))
        (print label)
        (push label labels)))))


(defun process()
  ; 'Main' function.
  ; Process each file path in turn and concatenate the lists of labels.
  ; Return all labels.
  (let ((file-list (get-files filepaths))
        (labels ()))
    (dolist (path file-list labels)
      (setf fp (openfile path))
      (setf labels (append labels (process-file fp)))
      (close fp))))


; Call the 'main' function, and catch any errors.
(catch 'err (process))
